import numpy as np
from collections import Counter
from scipy.sparse import csr_matrix
from scipy.sparse.csgraph import maximum_bipartite_matching
from scipy.optimize import linear_sum_assignment
from scipy.sparse.csgraph import min_weight_full_bipartite_matching


def search_minimum_sparsity(dist_mat, slackness, init_sparsity,
                            m_min, m_max, num_cells_to_use,
                            min_dist, verbose=True):
    """
    Use binary search to search for the minimum sparsity level k such that
    if dist_mat is trimmed to be a k-NN graph, a valid matching still exists.
    The search starts with k_left=1 and k_right=dist_mat.shape[0], and it is always true that:
    1) any k < k_left doesn't give a valid matching;
    2) any k >= k_right gives a valid matching.

    Parameters
    ----------
    dist_mat : array-like of shape (n_samples_1, n_samples_2)
        A two-dimensional distance matrix.
    slackness : int
        Binary search terminates when k_right - k_left <= slackness;
        an exact binary search corresponds to slackness = 0
    init_sparsity : int
        Binary search starts from k=init_sparsity.
    m_min : int
        Each row in the first dataset is matched to at least m_min many rows in the second dataset.
    m_max : int
        Each row in the first dataset is matched to at most m_max many rows in the second dataset.
    num_cells_to_use : int
        Total number of rows to use in the second dataset.
    min_dist : float
        It must be true that minimum entry in dist_mat is >= min_dist.

    Returns
    -------
    k_left : int
        If sparsity<k_left, then there is no valid matching.
    k_right : int
        If sparsity>=k_right, then there is a valid matching.
    """
    assert np.min(dist_mat) >= min_dist
    n1, n2 = dist_mat.shape

    num_sinks = max(n1 * m_max - num_cells_to_use, 0)

    # construct distance matrix that's ready for matching
    if m_max > 1:
        dist_mat = np.tile(dist_mat, (m_max, 1))

    argsort_res = np.argsort(dist_mat, axis=1)

    k_left = 1
    k_right = n2
    cnt = 0

    # start binary search
    while k_left < k_right - slackness:

        if verbose:
            print(
                'If sparsity>={}, then there is a valid matching; '
                'if sparsity<{}, then there is no valid matching.'.format(k_right, k_left), flush=True
            )
        if cnt == 0 and init_sparsity is not None:
            k = init_sparsity
        else:
            k = (k_left + k_right) // 2

        # construct k-NN graph from dist_mat
        # indices for the largest k entries
        largest_k = argsort_res[:, -(n2 - k):]
        # one means there is an edge and zero means there's no edge
        dist_bin = np.ones_like(dist_mat)
        dist_bin[np.arange(dist_mat.shape[0])[:, None], largest_k] = 0

        # add sinks if necessary
        if num_sinks > 0:
            # again, one means there is an edge,
            # and zero means there is no edge (i.e., the distance is np.inf)
            dist_sinks_bin = np.ones((dist_mat.shape[0], num_sinks))
            dist_sinks_bin[:m_min * n1, :] = 0
            dist_bin = np.concatenate((dist_bin, dist_sinks_bin), axis=1)

        dist_bin = csr_matrix(dist_bin)

        col_idx = maximum_bipartite_matching(dist_bin, perm_type='column')
        n_matched = np.sum(col_idx != -1)

        if n_matched == dist_bin.shape[0]:
            # the current k gives a valid matching
            # can trim more aggressively --> try a smaller k
            k_right = k
        else:
            # the current k doesn't give a valid matching
            # try a larger k
            k_left = k + 1

        cnt = cnt + 1

    # we know that k_right must give a valid matching
    if verbose:
        print(
            'If sparsity>={}, then there is a valid matching; '
            'if sparsity<{}, then there is no valid matching.'.format(k_right, k_left), flush=True
        )
    return k_left, k_right


def get_matching_from_indices(col_idx, n1, n2, m_max):
    """
    Assume col_idx is obtained from min_weight_full_bipartite_matching or linear_sum_assignment
    with some dist_mat as input.
    And this dist_mat is organized such that the sinks are in dist_mat[:, :n2].
    This function calculates the matching from col_idx.

    Parameters
    ----------
    col_idx : list
        output from min_weight_full_bipartite_matching or linear_sum_assignment.
    n1 : int
        Sample size of the first dataset.
    n2 : int
        Sample size of the second dataset.
    m_max : int
        Each row in the first dataset is matched to at most m_max many rows in the second dataset.

    Returns
    -------
    list
        A list of (potentially variable length) lists;
        it holds that the i-th row in the first dataset is matched to the res[i]-th row in the second dataset.
    """
    if m_max == 1:
        # make sure the result is a list of (length-one) lists
        return [[col_idx[ii]] for ii in range(n1)]

    res = {ii: [] for ii in range(n1)}

    for kk in range(m_max):
        for ii in range(n1):
            candidate = col_idx[ii + (kk - 1) * n1]
            if candidate < n2:
                res[ii].append(candidate)

    return [res[ii] for ii in range(n1)]


def match_cells(dist_mat, sparsity, m_min, m_max, num_cells_to_use,
                min_dist, mode='auto'):
    """
    Given dist_mat, first trim it to a k-NN graph according to the desired sparsity
    sparsity level, then do a matching according to the specified parameters.

    Parameters
    ----------
    dist_mat : array-like of shape (n_samples_1, n_samples_2)
        A two-dimensional distance matrix.
    sparsity : int
        An integer k such that dist_mat will be trimmed into a k-NN graph.
    m_min : int
        Each row in the first dataset is matched to at least m_min many rows in the second dataset.
    m_max : int
        Each row in the first dataset is matched to at most m_max many rows in the second dataset.
    num_cells_to_use : int
        Total number of samples to use in the second dataset.
    min_dist : float
        It must be true that minimum entry in dist_mat is >= min_dist.
    mode : str, default='auto'
        If 'sparse', use min_weight_full_bipartite_matching;
        if 'dense', use linear_sum_assignment;
        if 'auto': when sparsity<=n//2, use 'sparse', else use 'dense'.

    Returns
    -------
    list
        A list of (potentially variable length) lists;
        it holds that the i-th row in the first dataset is matched to the res[i]-th row in the second dataset.
    """
    assert np.min(dist_mat) >= min_dist
    n1, n2 = dist_mat.shape

    if m_max > 1:
        dist_mat = np.tile(dist_mat, (m_max, 1))

    num_sinks = max(n1 * m_max - num_cells_to_use, 0)

    if sparsity is None:
        mode = 'dense'
    elif mode == 'auto':
        mode = 'sparse' if sparsity <= n2 // 2 else 'dense'

    infinity_placeholder = 0 if mode == 'sparse' else np.Inf

    # trim nodes if necessary
    if sparsity is not None and sparsity != n2:
        argsort_res = np.argsort(dist_mat, axis=1)
        largest_k = argsort_res[:, -(n2 - sparsity):]
        # make a copy because some operations are in-place
        # and may modify the original dist_mat
        dist_mat_cp = np.copy(dist_mat)
        dist_mat_cp[np.arange(dist_mat.shape[0])[:, None], largest_k] = infinity_placeholder
    else:
        dist_mat_cp = dist_mat

    # add sinks if necessary
    if num_sinks > 0:
        # we need make sure that
        # those sinks are favored compared to all other nodes
        dist_sinks = np.zeros((dist_mat.shape[0], num_sinks)) + min_dist / 100
        dist_sinks[:m_min * n1, :] = infinity_placeholder
        dist_mat_cp = np.concatenate((dist_mat_cp, dist_sinks), axis=1)

    if mode == 'sparse':
        dist_mat_cp = csr_matrix(dist_mat_cp)
        _, col_idx = min_weight_full_bipartite_matching(dist_mat_cp)
    elif mode == 'dense':
        _, col_idx = linear_sum_assignment(dist_mat_cp)
    else:
        raise NotImplementedError

    return get_matching_from_indices(col_idx, n1, n2, m_max)


def is_perfect_matching(matching):
    """Check if matching is perfect (i.e., all nodes are matched) or not.
    Parameters
    ----------
    matching : list
        If matching[ii] is an empty list, then ii is not matched to anyone.

    Returns
    -------
    bool
        True if matching is perfect.
    """
    matching_len = np.array([len(matching[ii]) for ii in range(len(matching))])
    matched_flag = (matching_len > 0)
    return all(matched_flag)


def calculate_duplications(matching):
    """
    Calculate the number of i such that matching[i] appears more than one times.
    Also calculate the proportion of duplicated i's among all matched samples.

    Parameters
    ----------
    matching : list
        A list of (potentially variable length) lists;
        it holds that the i-th row in the first dataset is matched to the res[i]-th row in the second dataset.

    Returns
    -------
    dup_prop : float
         Proportion of i (among all matched samples) such that matching[i] appears more than one times.
    dup_freq : int
        Number of i such that matching[i] appears more than one times.
    """
    cells = []
    for matched in matching:
        cells.extend(matched)
    counter = Counter(cells)
    duplicated_cells = [cell for cell, freq in counter.items() if freq > 1]
    dup_prop = len(duplicated_cells) / len(cells)
    dup_freq = np.mean([counter[cell] for cell in duplicated_cells]) if len(duplicated_cells) > 0 else 0
    return dup_prop, dup_freq


def batching(df, num_batches):
    """
    Given a pandas dataframe, shuffle it and then cut it into approximately equal-sized batches.

    Parameters
    ----------
    df : pandas.DataFrame
        A pandas dataframe.
    num_batches : int
        Number of batches.

    Returns
    -------
    df_lst : list
        A list of pandas dataframe; each entry corresponds to one batch.
    perm_lst : list
        A list of nested lists of indices; perm_lst[i] corresponds to the original row indices of df_lst[i].
    """
    n = df.shape[0]
    perm = np.random.permutation(n)
    size_each, remainder = divmod(n, num_batches)
    size_last_batch = size_each + remainder
    start_idx = 0
    df_lst = []
    perm_lst = []
    for i in range(num_batches):
        end_idx = start_idx + size_last_batch if i == num_batches-1 else start_idx + size_each
        indices = perm[start_idx:end_idx]
        perm_lst.append(indices)
        df_lst.append(df.iloc[indices])
        start_idx = end_idx
    return df_lst, perm_lst


def stitch(perm_lst, matching_lst, n):
    """Stitch multiple matching into one.

    Parameters
    ----------
    perm_lst : list
        Output from batching.
    matching_lst : list
        perm_lst[i][j] should be matched to matching_lst[i][j].
    n : int
        Sample size.

    Returns
    -------
    list
        A list of (potentially variable length) lists;
        it holds that the i-th row in the first dataset is matched to the res[i]-th row in the second dataset.
    """
    stitched_dict = {} # store perm_lst[i][j]: matching_lst[i][j]
    for i in range(len(perm_lst)):
        for j in range(len(perm_lst[i])):
            stitched_dict[perm_lst[i][j]] = matching_lst[i][j]
    return [stitched_dict[i] for i in range(n)]


def eval_matching_accuracy(X_labels, Y_labels, matching, criterion='maj'):
    """Evaluate matching accuracy.

    Parameters
    ----------
    X_labels : array-like of shape (n_samples_of_X)
        The cluster labels of the first dataset.
    Y_labels : array-like of shape (n_samples_of_Y)
        The cluster labels of the second dataset.
    matching : list
        A list of (potentially variable length) lists;
        it holds that the i-th row in the first dataset is matched to the res[i]-th row in the second dataset.
    criterion : str, default='maj'
        If 'any': correct if X_labels[i] == any of Y_labels[i];
        if 'all': correct if X_labels[i] == all of Y_labels[i];
        if 'maj': correct if X_labels[i] == the mode in Y_labels[i], with ties broken in an arbitrary fashion.

    Returns
    -------
    float
        The proportion of correctly matched cells in X.
    """
    X_labels = np.array(X_labels)
    Y_labels = np.array(Y_labels)
    n = len(X_labels)
    assert n == len(matching)

    # the matching is of variable length
    acc = 0
    cnt = 0
    for ii in range(n):
        if len(matching[ii]) == 0:
            continue
        elif len(matching[ii]) == 1:
            cnt = cnt + 1
            acc = acc + int(X_labels[ii] == Y_labels[matching[ii][0]])
        else:
            cnt = cnt + 1
            flag = [X_labels[ii] == y for y in Y_labels[matching[ii]]]
            if criterion == 'any':
                acc = acc + any(flag)
            elif criterion == 'all':
                acc = acc + all(flag)
            elif criterion == 'maj':
                acc = acc + (np.sum(flag) > (len(flag) // 2))
            else:
                raise NotImplementedError

    return acc / cnt


import numpy as np
import pandas as pd
import warnings
import os
from sklearn.utils.extmath import randomized_svd
from . import match_utils, utils, embed
from .cluster import spectral_clustering, jr_kmeans


class Mario(object):
    def __init__(self, df1, df2, normalization=True):
        """Initialize the Mario object.

        Parameters
        ----------
        df1 : array-like of shape (n_samples_1, n_features_1)
            The first dataset.
        df2 : array-like of shape (n_samples_2, n_features_2)
            The second dataset.
        normalization : bool, default=True
            If true, center each column and scale each column to have unit standard deviation.
        """
        # convert df1 and df2 to dataframe if they are not
        if not isinstance(df1, pd.DataFrame):
            df1 = pd.DataFrame(df1)
        if not isinstance(df2, pd.DataFrame):
            df2 = pd.DataFrame(df2)

        self.min_dist = 1e-5
        # parameters related to datasets
        if normalization:
            self.df1 = utils.normalize(df1)
            self.df2 = utils.normalize(df2)
        else:
            self.df1 = df1
            self.df2 = df2
        self.n1, self.p1 = df1.shape
        self.n2, self.p2 = df2.shape
        assert self.n1 <= self.n2
        self.ovlp_features = [x for x in df1.columns if x in df2.columns]

        # hyper-parameters
        self.n_matched_per_cell = None
        self.m_min = None
        self.m_max = None
        self.num_cells_to_use = None
        self.num_sinks = None
        self.sparsity = {'ovlp': None, 'all': None}
        self.n_components = {'ovlp': None, 'all': None}

        # cache some results
        self.dist = {'ovlp': None, 'all': None}
        self.matching = {'ovlp': None, 'all': None, 'wted': None, 'final': None, 'knn': None}
        self.best_wt = None
        self.stacked_svd = {'U': None, 's': None, 'Vh': None}
        self.ovlp_cancor = None
        self.ovlp_scores = {'x': None, 'y': None}

    def compute_dist_ovlp(self, n_components=10):
        """Compute distance matrix based on overlapping features.

        Parameters
        ----------
        n_components : int
            Number of SVD components to keep

        Returns
        -------
        dist_ovlp : array-like of shape (n1, n2)
            The distance matrix based on the overlapping features.
        s : array-like of shape (n_components, )
            Vector of singular values.
        """
        if n_components > len(self.ovlp_features):
            warnings.warn("n_components exceed the number of overlapping features,"
                          " set it to be the number of overlapping features.")
            n_components = len(self.ovlp_features)

        self.n_components['ovlp'] = n_components

        if not (
                self.stacked_svd['U'] is not None and self.stacked_svd['s'] is not None
                and self.stacked_svd['Vh'] is not None and len(self.stacked_svd['s']) >= n_components
        ):
            # Cached results are not valid, do SVD
            arr1 = utils.normalize(self.df1[self.ovlp_features]).to_numpy()
            arr2 = utils.normalize(self.df2[self.ovlp_features]).to_numpy()

            self.stacked_svd['U'], self.stacked_svd['s'], self.stacked_svd['Vh'] = \
                randomized_svd(np.concatenate((arr1, arr2), axis=0), n_components=n_components)
            if n_components == len(self.ovlp_features):
                dist_mat = utils.cdist_correlation(arr1, arr2)
            else:
                svd1 = self.stacked_svd['U'][:self.n1, :] @ np.diag(self.stacked_svd['s']) @ self.stacked_svd['Vh']
                svd2 = self.stacked_svd['U'][self.n1:, :] @ np.diag(self.stacked_svd['s']) @ self.stacked_svd['Vh']
                dist_mat = utils.cdist_correlation(svd1, svd2)
        else:
            # use cached results
            svd1 = self.stacked_svd['U'][:self.n1, :n_components] @ np.diag(self.stacked_svd['s'][:n_components]) \
                   @ self.stacked_svd['Vh'][:n_components, :]
            svd2 = self.stacked_svd['U'][self.n1:, :n_components] @ np.diag(self.stacked_svd['s'][:n_components]) \
                   @ self.stacked_svd['Vh'][:n_components, :]
            dist_mat = utils.cdist_correlation(svd1, svd2)

        # make sure min_dist is at least self.min_dist
        dist_mat = utils.check_min_dist(dist_mat, self.min_dist)
        self.dist['ovlp'] = dist_mat
        return dist_mat, self.stacked_svd['s'][:n_components]

    def search_minimum_sparsity(self, dist_mat, slackness=200, init_sparsity=None, verbose=True):
        """
        Use binary search to search for the minimum sparsity level k such that
        if dist_mat is trimmed to be a k-NN graph, a valid matching still exists.
        The search starts with k_left=1 and k_right=dist_mat.shape[0], and it is always true that:
        1) any k < k_left doesn't give a valid matching;
        2) any k >= k_right gives a valid matching.

        Parameters
        ----------
        dist_mat : array-like of shape (n_samples_1, n_samples_2)
            A two-dimensional distance matrix.
        slackness : int, default=200
            Binary search terminates when k_right - k_left <= slackness;
            an exact binary search corresponds to slackness = 0
        init_sparsity : int, default=None
            Binary search starts from k=init_sparsity. If None, start from the middle.
        verbose : bool, default=None
            If True, print the progress.

        Returns
        -------
        k_left : int
            If sparsity<k_left, then there is no valid matching.
        k_right : int
            If sparsity>=k_right, then there is a valid matching.
        """
        return match_utils.search_minimum_sparsity(
            dist_mat, slackness, init_sparsity, self.m_min,
            self.m_max, self.num_cells_to_use, self.min_dist, verbose
        )

    def specify_matching_params(self, n_matched_per_cell):
        """Specify how many cells in the second dataset are to be matched with one cell in the first dataset.

        Parameters
        ----------
        n_matched_per_cell : int
            How many cells in the second dataset are to be matched with one cell in the first dataset.
        """
        self.n_matched_per_cell = n_matched_per_cell
        if self.n1 * n_matched_per_cell > self.n2:
            raise ValueError("Not enough cells in Y data!")
        self._specify_matching_params(1, n_matched_per_cell, self.n1 * n_matched_per_cell)

    def _specify_matching_params(self, m_min, m_max, num_cells_to_use):
        """Specify the matching parameters.

        Parameters
        ----------
        m_min : int
            Each row in the first dataset is matched to at least m_min many rows in the second dataset.
        m_max : int
            Each row in the first dataset is matched to at most m_max many rows in the second dataset.
        num_cells_to_use : int
            Total number of rows to use in the second dataset.
        -------
        """
        assert m_min >= 1
        assert self.n1 <= num_cells_to_use
        assert num_cells_to_use <= self.n1 * m_max
        assert num_cells_to_use <= self.n2
        # m_max cannot be too large, otherwise a matching does not exist
        assert m_max <= num_cells_to_use - (self.n1 - 1)

        self.m_min = m_min
        self.m_max = m_max
        self.num_cells_to_use = num_cells_to_use
        self.num_sinks = max(self.n1 * m_max - num_cells_to_use, 0)

    def match_cells(self, dist_mat='ovlp', sparsity=None, mode='auto'):
        """Do cell matching.

        Parameters
        ----------
        dist_mat : str or a user-specified distance matrix, default='ovlp'
            If 'ovlp', then match using the distance matrix computed from overlapping features;
            if 'all', then match using the distance matrix computed from all the features;
            if a user-specified array-like of shape (n1, n2), then match using this distance matrix.
        sparsity : int
            Number of nearest neighbors to keep in the distance matrix.
        mode : str, default='auto'
            If 'sparse', use min_weight_full_bipartite_matching;
            if 'dense', use linear_sum_assignment;
            if 'auto': when sparsity<=n//2, use 'sparse', else use 'dense'.

        Returns
        -------
        list
            A list of (potentially variable length) lists;
            it holds that the i-th row in the first dataset is matched to the res[i]-th row in the second dataset.
        """
        if isinstance(dist_mat, str):
            if dist_mat not in self.dist or self.dist[dist_mat] is None:
                raise ValueError("Distance not found!")
            self.sparsity[dist_mat] = sparsity
            try:
                self.matching[dist_mat] = match_utils.match_cells(
                    self.dist[dist_mat], sparsity, self.m_min, self.m_max, self.num_cells_to_use, self.min_dist, mode
                )
            except ValueError:
                # too sparse, find the suitable sparsity level
                warnings.warn(
                    'Current sparsity config '
                    'is too sparse, finding a suitable sparsity level...'
                )
                _, new_sparsity = self.search_minimum_sparsity(
                    self.dist[dist_mat], slackness=200, init_sparsity=self.sparsity[dist_mat] + 1, verbose=True
                )
                warnings.warn('The new sparsity level is {}'.format(new_sparsity))
                self.sparsity[dist_mat] = new_sparsity
                self.matching[dist_mat] = match_utils.match_cells(
                    self.dist[dist_mat], new_sparsity, self.m_min, self.m_max, self.num_cells_to_use, self.min_dist,
                    mode
                )
            return self.matching[dist_mat]
        else:
            try:
                matching = match_utils.match_cells(
                    dist_mat, sparsity, self.m_min, self.m_max, self.num_cells_to_use, self.min_dist, mode
                )
            except ValueError:
                # too sparse, find the suitable sparsity level
                warnings.warn(
                    'Current sparsity config '
                    'is too sparse, finding a suitable sparsity level...'
                )
                _, new_sparsity = self.search_minimum_sparsity(
                    self.dist[dist_mat], slackness=200, init_sparsity=sparsity + 1, verbose=True
                )
                warnings.warn('The new sparsity level is {}'.format(new_sparsity))
                matching = match_utils.match_cells(
                    dist_mat, new_sparsity, self.m_min, self.m_max, self.num_cells_to_use, self.min_dist, mode
                )
            return matching

    def _align_modalities(self, matching):
        """Align df1 so so that cell i in df1 is matched to the "averaged cell" in cells matching[i] in df2.

        Parameters
        ----------
        matching : list
            A list of (potentially variable length) lists;
            it holds that the i-th row in the first dataset is matched to the res[i]-th row in the second dataset.

        Returns
        -------
        X : array-like of shape (n_samples, n_features_1)
            The first dataset.
        Y : array-like of shape (n_samples, n_features_2)
            The second dataset after alignment.
        """
        assert len(matching) == self.n1

        # if cell ii in df1 is filtered out, then matching[ii] is an empty list
        X = []
        Y = []
        for ii in range(self.n1):
            if len(matching[ii]) == 0:
                continue

            X.append(self.df1.iloc[ii, :])
            if len(matching[ii]) == 1:
                Y.append(self.df2.iloc[matching[ii][0]])
            else:
                Y.append(self.df2.iloc[matching[ii]].mean(axis=0))

        X = np.array(X)
        Y = np.array(Y)

        return X, Y

    def fit_cca(self, matching, n_components=20, max_iter=3000):
        """Align df1 and df2 using matching, then fit a CCA.

        Parameters
        ----------
        matching : str or list
            Either 'ovlp', 'all', 'wted', 'final', or 'knn',
            or a list of (potentially variable length) lists;
            it holds that the i-th row in the first dataset is matched to the res[i]-th row in the second dataset.
        n_components : int
            Number of components for CCA.
        max_iter : int
            Maximum iteration for CCA.

        Returns
        -------
        cancor: array-like of shape (n_components, )
            Vector of canonical components.
        cca: CCA
            CCA object.
        """
        if n_components > min(self.p1, self.p2):
            warnings.warn('n_components must be <= the dimensions of the two datasets, '
                          'set it to be equal to the minimum of the dimensions of the two datasets')
            n_components = min(self.p1, self.p2)
        if isinstance(matching, str):
            X, Y = self._align_modalities(self.matching[matching])
        else:
            X, Y = self._align_modalities(matching)
        cancor, cca = embed.get_cancor(X, Y, n_components, max_iter)

        return cancor, cca

    def compute_dist_all(self, matching='ovlp', n_components=20, max_iter=5000):
        """Given matching, align df1 and df2, fit a CCA, then use CCA scores to get the distance matrix.

        Parameters
        ----------
        matching : str or list
            Either 'ovlp', meaning that we use the matching obtained from the overlapping features;
            or a list of (potentially variable length) lists, cell i in df1 is matched to cell matching[i] in df2.
        n_components : int
            Number of CCA components.
        max_iter : int
            Max number of iterations when fitting CCA.

        Returns
        -------
        dist_mat : array-like of shape (n1, n2)
            The distance matrix.
        cancor: array-like of shape (n_components, )
            Vector of canonical components.
        """
        if n_components <= 1:
            n_components = 2
            warnings.warn('n_components must be at least 2, '
                          'set it to 2')

        if n_components > min(self.p1, self.p2):
            warnings.warn('n_components must be <= the dimensions of the two datasets, '
                          'set it to be equal to the minimum of the dimensions of the two datasets')
            n_components = min(self.p1, self.p2)

        self.n_components['all'] = n_components

        if isinstance(matching, str) and matching == 'ovlp':
            if self.matching['ovlp'] is None:
                raise ValueError('Initial matching not found!')
            if not (
                    self.ovlp_scores['x'] is not None and self.ovlp_scores['y'] is not None
                    and self.ovlp_scores['x'].shape[1] >= n_components
            ):
                # cached results are not valid, do CCA
                cancor, cca = self.fit_cca(self.matching['ovlp'], n_components, max_iter)
                self.ovlp_cancor = cancor
                self.ovlp_scores['x'], self.ovlp_scores['y'] = cca.transform(self.df1, self.df2)
                dist_mat = utils.cdist_correlation(self.ovlp_scores['x'], self.ovlp_scores['y'])
            else:
                # use cached results
                dist_mat = utils.cdist_correlation(self.ovlp_scores['x'][:, :n_components],
                                                   self.ovlp_scores['y'][:, :n_components])
                cancor = self.ovlp_cancor[:n_components]
        else:
            # use user-specified matching
            cancor, cca = self.fit_cca(matching, n_components, max_iter)
            df1_cca, df2_cca = cca.transform(self.df1, self.df2)
            dist_mat = utils.cdist_correlation(df1_cca, df2_cca)

        self.dist['all'] = utils.check_min_dist(dist_mat, self.min_dist)
        return self.dist['all'], cancor

    def _matchable(self, n_sim=20, top_k=10, flip_prob=0.3, subsample_prop=1, verbose=True):
        """Calculate the p-value for the hypothesis that two datasets are matchable

        Parameters
        ----------
        n_sim : int, default=20
            Number of simulation rounds.
        top_k : int, default=10
            The mean of the top_k canonical correlations will be taken as the test statistic.
        flip_prob : float, default=0.3
            Probability of sign flip.
        subsample_prop : float, default=1
            Random subsample df1 and df2 to speed up the computation, 1 means using the full data.
        verbose : bool, default=True
            Print details if True.

        Returns
        -------
        pval_ovlp : float
            The p-value for the hypothesis that two datasets are matchable based on overlapping features.
        pval_all : float
            The p-value for the hypothesis that two datasets are matchable based on all the features.
        """
        cancor_ovlp_list = []
        cancor_all_list = []
        if self.n_components['ovlp'] is None or self.n_components['all'] is None:
            raise ValueError("Please do both initial and refined matching first before calling this function.")

        if subsample_prop < 1:
            df1_subsampled = self.df1.iloc[np.random.choice(self.n1, int(self.n1 * subsample_prop), replace=False)]
            df2_subsampled = self.df2.iloc[np.random.choice(self.n2, int(self.n2 * subsample_prop), replace=False)]
        else:
            df1_subsampled = self.df1
            df2_subsampled = self.df2

        match_subsampled = Mario(df1_subsampled, df2_subsampled, normalization=True)
        match_subsampled.specify_matching_params(self.n_matched_per_cell)
        # do matching
        _ = match_subsampled.compute_dist_ovlp(self.n_components['ovlp'])
        _ = match_subsampled.match_cells('ovlp', self.sparsity['ovlp'], mode='auto')

        # use all features
        _ = match_subsampled.compute_dist_all('ovlp', n_components=self.n_components['all'])
        _ = match_subsampled.match_cells('all', self.sparsity['all'], mode='auto')

        # calculate test statistics
        cancor_ovlp_obs = np.mean(match_subsampled.fit_cca(match_subsampled.matching['ovlp'], n_components=top_k)[0])
        cancor_all_obs = np.mean(match_subsampled.fit_cca(match_subsampled.matching['all'], n_components=top_k)[0])

        for ii in range(n_sim):
            if verbose:
                print("Random sign flip, round {}...".format(ii), flush=True)
            max_trial = 100
            # looks like we may get SVD not converge error
            # and it is due to linear algebra libraries in the linux server
            # so try max_trial times until success
            trial_idx = 0
            while trial_idx < max_trial:
                trial_idx += 1
                try:
                    rand_signs_1 = 2 * np.random.binomial(1, 1 - flip_prob, match_subsampled.n1) - 1
                    rand_signs_2 = 2 * np.random.binomial(1, 1 - flip_prob, match_subsampled.n2) - 1
                    df1_flipped = (match_subsampled.df1.T * rand_signs_1).T
                    df2_flipped = (match_subsampled.df2.T * rand_signs_2).T

                    match_flipped = Mario(df1_flipped, df2_flipped, normalization=False)
                    match_flipped.specify_matching_params(self.n_matched_per_cell)

                    # use ovlp features
                    _ = match_flipped.compute_dist_ovlp(self.n_components['ovlp'])
                    _ = match_flipped.match_cells('ovlp', self.sparsity['ovlp'], mode='auto')

                    # calculate the median/mean of top_k cancors
                    cancor_ovlp_list.append(np.mean(
                        match_flipped.fit_cca(match_flipped.matching['ovlp'], n_components=top_k)[0]
                    ))

                    # use all features
                    _ = match_flipped.compute_dist_all('ovlp', n_components=self.n_components['all'])
                    _ = match_flipped.match_cells('all', self.sparsity['all'], mode='auto')

                    # calculate the median/mean of top_k cancors
                    cancor_all_list.append(np.mean(
                        match_flipped.fit_cca(match_flipped.matching['all'], n_components=top_k)[0]
                    ))
                except Exception as exception:
                    Warning("{} occurs at the {}/{}-th trial, try flipping again...".format(
                        type(exception).__name__, trial_idx, max_trial)
                    )
                    continue
                else:
                    # the code will arrive here only if try has succeeded
                    break
            if trial_idx == max_trial:
                raise ValueError("{} consecutive trials have all failed.".format(max_trial))
            elif trial_idx > 1:
                warnings.warn("Success at {}/{}-th trial".format(trial_idx, max_trial))

        # calculate p-values
        pval_ovlp = np.mean([cancor_ovlp_list[ii] >= cancor_ovlp_obs for ii in range(n_sim)])
        pval_all = np.mean([cancor_all_list[ii] >= cancor_all_obs for ii in range(n_sim)])
        return pval_ovlp, pval_all

    def matchable(self, n_sim=20, top_k=10, flip_prob=0.3, subsample_prop=1, subsample_rounds=1, verbose=True):
        """Calculate the p-value for the hypothesis that two datasets are matchable.

        Parameters
        ----------
        n_sim : int, default=20
            Number of simulation rounds.
        top_k : int, default=10
            The mean of the top_k canonical correlations will be taken as the test statistic.
        flip_prob : float, default=0.3
            Probability of sign flip.
        subsample_prop : float, default=1
            Subsample to speed up computation, 1 means using the full data.
        subsample_rounds : int, default=1
            The subsampling operation is done this many rounds and the p-values are averaged.
        verbose : bool, default=True
            Print details if True.

        Returns
        -------
        pval_ovlp : float
            The p-value for the hypothesis that two datasets are matchable based on overlapping features.
        pval_all : float
            The p-value for the hypothesis that two datasets are matchable based on all the features.
        """
        pval_ovlp = 0
        pval_all = 0
        for i in range(subsample_rounds):
            if verbose and subsample_prop < 1:
                print(f"Now at subsample round {i}...", flush=True)
            curr_pval_ovlp, curr_pval_all = self._matchable(n_sim, top_k, flip_prob, subsample_prop, verbose)
            pval_ovlp += curr_pval_ovlp
            pval_all += curr_pval_all

        return pval_ovlp / subsample_rounds, pval_all / subsample_rounds

    def interpolate(self, n_wts=10, top_k=10, verbose=True):
        """
        Let wt_vec be an evenly spaced list from 0 to 1 with length n_wts.
        For each wt in wt_vec, do matching on (1-wt)*dist_ovlp + wt*dist_all,
        and select the best wt according to the mean of top_k canonical correlations.

        Parameters
        ----------
        n_wts : int, default=10
            wt_vec is a evenly spaced list from 0 to 1 with length n_wts.
        top_k : int, default=10
            The mean of top_k canonical correlations is taken as the quality measure.
        verbose : bool, default=True
            Print details if True.

        Returns
        -------
        best_wt : float
            The best wt in wt_vec.
        best_matching : list
            The matching corresponds to best_wt.
        """
        wt_vec = np.linspace(0, 1, n_wts)
        max_cancor = float('-inf')
        for ii in range(n_wts):
            if verbose:
                print('Now at iteration {}, wt={}'.format(ii, wt_vec[ii]), flush=True)
            if ii == 0:
                # curr_dist = self.dist['ovlp']
                curr_matching = self.matching['ovlp']
            elif ii == n_wts - 1:
                # curr_dist = self.dist['all']
                curr_matching = self.matching['all']
            else:
                # ii small --> more close to dist_ovlp
                curr_dist = (1 - wt_vec[ii]) * self.dist['ovlp'] + wt_vec[ii] * self.dist['all']
                if self.sparsity['ovlp'] is None or self.sparsity['all'] is None:
                    curr_sparsity = self.sparsity['ovlp'] or self.sparsity['all']
                else:
                    curr_sparsity = max(self.sparsity['ovlp'], self.sparsity['all'])
                try:
                    curr_matching = self.match_cells(curr_dist, curr_sparsity, 'auto')
                except ValueError:
                    # too sparse, find the suitable sparsity level
                    if verbose:
                        print(
                            'Current sparsity config '
                            'is too sparse, finding a suitable sparsity level...', flush=True
                        )
                    _, curr_sparsity = self.search_minimum_sparsity(
                        curr_dist, slackness=200, init_sparsity=curr_sparsity + 1, verbose=verbose
                    )
                    curr_matching = self.match_cells(curr_dist, curr_sparsity, mode='auto')

            # compute median/mean of top_k cancors
            curr_cancor = np.mean(self.fit_cca(curr_matching, n_components=top_k)[0])
            if curr_cancor > max_cancor:
                max_cancor = curr_cancor
                self.matching['wted'] = curr_matching
                self.best_wt = wt_vec[ii]

        return self.best_wt, self.matching['wted']

    def knn_matching(self, dist_mat='wted', k=5):
        """Matching via k-nearest neighbors.

        Parameters
        ----------
        dist_mat : str or a user-specified distance matrix, default='wted'
            If 'ovlp', then match using the distance matrix computed from overlapping features;
            if 'all', then match using the distance matrix computed from all the features;
            if 'wted', then using the best interpolated distance matrix;
            if a user-specified array-like of shape (n1, n2), then match using this distance matrix.
        k : int
            Number of nearest neighbors.

        Returns
        -------
        list
            A list of (potentially variable length) lists;
            it holds that the i-th row in the first dataset is matched to the res[i]-th row in the second dataset.
        """
        if dist_mat == 'wted':
            assert self.best_wt is not None
            dist_argsort = np.argsort((1 - self.best_wt) * self.dist['ovlp'] + self.best_wt * self.dist['all'], axis=1)
        elif dist_mat == 'ovlp':
            dist_argsort = np.argsort(self.dist['ovlp'], axis=1)
        elif dist_mat == 'all':
            dist_argsort = np.argsort(self.dist['all'], axis=1)
        else:
            dist_argsort = np.argsort(dist_mat, axis=1)

        dist_argsort = dist_argsort[:, :k]
        self.matching['knn'] = [dist_argsort[ii, :].tolist() for ii in range(self.n1)]
        return self.matching['knn']

    def _jr_clustering(self, matching, n_clusters, n_components=20, mismatch_prob=0.1, max_iter=50, tol=1e-5,
                       verbose=True):
        """
        Given matching:
        1) align df1 and df2, get df1_aligned and df2_aligned;
        2) do CCA on df1_aligned and df2_aligned to get CCA scores, call them df1_cca and df2_cca;
        3) do spectral clustering on the averaged df, i.e., (df1_cca+df2_cca)/2 to get init_labels;
        4) given init_labels, run jr_kmeans on df1_aligned and df2_aligned.

        Parameters
        ----------
        matching : list
            Matching used to align df1 and df2, must be a perfect matching
        n_clusters : int
            Number of clusters desired.
        n_components : int, default=20
            Number of CCA components for initial clustering.
        mismatch_prob : float, default=0.1
            Approximately mismatch_prob proportion of cells have different cluster labels.
        max_iter : int, default=50
            Max iteration desired.
        tol : float, default=1e-5.
            If the objective function changes less than tol, terminates jr_kmeans.
        verbose : bool, default=True
            If True, print details when running jr_kmeans.

        Returns
        -------
        cluster_labels : list
            cluster_labels[0] (resp. [1]) is the cluster labels for df1_aligned (resp. df2_aligned).
        centroids : list
            centroids[0] (resp. [1]) is an array of shape (n_clusters, df1.shape[1])
            (resp. of shape (n_clusters, df2.shape[1])) representing the centroids for df1_aligned (resp. df2_aligned).
        """

        assert match_utils.is_perfect_matching(matching)

        df1_aligned, df2_aligned = self._align_modalities(matching)
        _, cca = embed.get_cancor(df1_aligned, df2_aligned, n_components)
        df1_cca, df2_cca = cca.transform(df1_aligned, df2_aligned)
        # df1_cca = df1_cca / np.std(df1_cca, axis=0)
        # df2_cca = df2_cca / np.std(df2_cca, axis=0)
        df1_cca = utils.normalize(df1_cca)
        df2_cca = utils.normalize(df2_cca)
        init_cluster_labels = spectral_clustering(
            (df1_cca + df2_cca) / 2,
            n_clusters=n_clusters,
            need_svd=True
        )

        cluster_labels, _, centroids = jr_kmeans(
            [df1_aligned, df2_aligned],
            [init_cluster_labels, init_cluster_labels],
            n_clusters, mismatch_prob, max_iter, tol, verbose
        )

        return cluster_labels, centroids

    def filter_bad_matches(self, matching='wted', n_clusters=15,
                           n_components=20, bad_prop=0.1, max_iter=50,
                           tol=1e-5, verbose=True):
        """
        Given matching:
        1) run self.jr_clustering to get cluster_labels,
           note that cluster_labels[0] is the cluster labels for df1_aligned,
           and cluster_label[1] is the cluster labels for df2_aligned
        2) matching[ii] is bad if cluster_labels[0][ii] != cluster_labels[1][ii],
           then set matching[ii] to be an empty list

        Parameters
        ----------
        matching : str or list, default='wted'
            A matching to be filtered, must be a perfect matching.
            If 'ovlp', 'all', or 'wted', then use the matching obtained from
            the overlapping features, all the features, and their interpolation, respectively.
            Otherwise, a list corresponds to user-specified matching.
        n_clusters : int, default=15,
            How many clusters desired in self.jr_clustering.
        n_components : int, default=20
            Number of CCA components in the initialization step in self.jr_clustering.
        bad_prop : float, default=0.1
            Approximate proportion of bad matches to be filtered out.
        max_iter : int, default=50
            Run self.jr_clustering for how many iterations.
        tol : int, default=1e-5
            Terminate self.jr_clustering when the change of objective function is <= tol.
        verbose : bool, default=True
            Print all the details when True.

        Returns
        -------
        good_matching : list
            If for some i, matching[i] is an empty list, it means that i is bad and has been filtered out.
        """

        # proportion of good cells is approximately (1-mismatch_prob)^2 + mismatch_prob^2/(n_clusters-1),
        # this is because the model says that a cluster label flips w.p. mismatch_prob
        # and we declare a pair to be good if either (1) both not flipped, or (2) both flipped
        # so given bad_prop, can solve for the corresponding mismatch_prob

        assert 1e-10 < bad_prop < 0.5 - 1e-10

        if n_components > min(self.p1, self.p2):
            warnings.warn('n_components must be <= the dimensions of the two datasets, '
                          'set it to be equal to the minimum of the dimensions of the two datasets')
            n_components = min(self.p1, self.p2)

        mismatch_prob = (1 - np.sqrt(1 - 2*bad_prop*(1+1/(n_clusters-1)))) / (1 + 1/(n_clusters-1))

        if isinstance(matching, str):
            if matching not in self.matching or self.matching[matching] is None:
                raise ValueError("Matching not found.")
            matching = self.matching[matching]

        cluster_labels, _ = self._jr_clustering(
            matching, n_clusters, n_components, mismatch_prob,
            max_iter, tol, verbose
        )

        good_matching = [x for x in matching]

        for ii in range(self.n1):
            if cluster_labels[0][ii] != cluster_labels[1][ii]:
                good_matching[ii] = []

        self.matching['final'] = good_matching

        return self.matching['final']


def pipelined_mario(data_lst, normalization=True, n_batches=4,
                    n_matched_per_cell=1, sparsity_ovlp=None, sparsity_all=None,
                    n_components_ovlp=20, n_components_all=20,
                    n_cancor=5, n_wts=10,
                    n_clusters=10, n_components_filter=10, bad_prop=0.1, max_iter_filter=20,
                    knn=False, embed_dim=20, normalize_embed=True, max_iter_embed=500, save_path='.', verbose=False):
    """Run the whole Mario pipeline.

    Parameters
    ----------
    data_lst : list
        A list of array-likes, the i-th entry is an array-like of shape (ni, pi);
        will match data_lst[0] with data_lst[i] for i>=1.
    normalization : bool, default=True
        If true, center each column and scale each column to have unit standard deviation.
    n_batches : int, default=4
        Randomly cut data_list[0] into batches and match each batch to the rest of datasets.
    n_matched_per_cell : int
        How many cells in the second dataset are to be matched with one cell in the first dataset.
    sparsity_ovlp : int or None
        Sparsity level for matching with overlapping features.
    sparsity_all : int or None
        Sparsity level for matching with all features.
    n_components_ovlp : int, default=20
        Number of SVD components when calculating the distance matrix using overlapping features.
    n_components_all : int, default=20
        Number of CCA components when calculating the distance matrix using all the features.
    n_cancor : int, default=5
        Use the mean of top canonical correlations to choose the best interpolated matching.
    n_wts : int, default=10
        True n_wts-many evenly-spaced weights.
    n_clusters : int, default=10
        Number of clusters when doing regularized filtering.
    n_components_filter : int, default=10
        Number of SVD components when doing regularized filtering.
    bad_prop : float, default=0.1
        Approximate proportion of bad matches to be filtered out.
    max_iter_filter : int default=20
        Maximum iteration when doing filtering.
    knn : bool or int, default=False
        If is an integer k, run k-nn matching.
    embed_dim : int, default=20
        Number of CCA components when calculating joint embeddings.
    normalize_embed : bool, default=True
        Whether to normalize (center and scale to have unit standard deviation) the CCA scores.
    max_iter_embed : int, default=1000
        Number of maximum GCCA iterations.
    save_path : str, default='.'
        The path to save the results, will be created if not exist.
        final_matching and knn_matching will be stored in arrays of shape (n0, len(data_lst)),
        where the first column is the original row indices of data_lst[0] (i.e., 0, 1, ..., n1-1),
        the rest of columns are the matched cells in data_list[i] for i>=1.
        For example, if there are three datasets, then the first row could be '0', '1,5,20', '1,5,21',
        meaning that the first cell in data_lst[0] is matched to cells 1, 5, and 20 in data_lst[1]
        and is matched to cells 1, 5, 21 in data_lst[2].
        The arrays corresponding to final_matching and knn_matching (if knn is not False) will be saved at
        'save_path/final_matching.csv' and 'save_path/knn_matching.csv', respectively.
        The embeddings will be saved at 'save_path/embedding_i.csv' for 0 <= i < len(data_lst).
    verbose : bool, default=True
        Print the details if True.

    Returns
    -------
    final_matching_lst : list
        A list of length len(data_lst), whose i-th entry is the final matching (after interpolation and filtering)
        from data_lst[0] to data_lst[i]. The first entry is [[0], [1], ..., [n0-1]].
    knn_matching_lst : list
        A list of length len(data_lst), whose i-th entry is the knn matching
        from data_lst[0] to data_lst[i]. The first entry is [[0], [1], ..., [n0-1]].
        When knn=False, this list is not returned.
    embedding_lst : list
        A list of length len(data_lst), whose i-th entry is an array-like of shape (ni, embed_dim).
    """
    # convert datasets into dataframe if they are not
    for i in range(len(data_lst)):
        if not isinstance(data_lst[i], pd.DataFrame):
            data_lst[i] = pd.DataFrame(data_lst[i])
        if normalization:
            data_lst[i] = utils.normalize(data_lst[i])

    # cut data_lst[0] into batches
    df1 = data_lst[0]
    df1_lst, perm_lst = match_utils.batching(df1, n_batches)

    final_matching_lst = [[[i] for i in range(df1.shape[0])]]
    do_knn = (not isinstance(knn, bool)) and isinstance(knn, int) and knn >= 1
    if do_knn:
        knn_matching_lst = [[[i] for i in range(df1.shape[0])]]

    for i in range(1, len(data_lst)):
        print("Matching data_lst[0] with data_lst[{}]".format(i), flush=True)
        df2 = data_lst[i]
        matching_per_batch = []
        if do_knn:
            wted_dist_lst = []  # calculate weighted distance for later usage
        for j in range(n_batches):
            print('Now at batch {}'.format(j), flush=True)
            mario = Mario(df1_lst[j], df2, normalization=False)
            print('Matching using overlapping features...', flush=True)
            mario.specify_matching_params(n_matched_per_cell)
            _ = mario.compute_dist_ovlp(n_components_ovlp)
            _ = mario.match_cells('ovlp', sparsity=sparsity_ovlp)
            print('Matching using all features...', flush=True)
            _ = mario.compute_dist_all('ovlp', n_components_all)
            _ = mario.match_cells('all', sparsity=sparsity_all)
            print('Finding the best interpolated matching...', flush=True)
            wt, _ = mario.interpolate(n_wts, n_cancor, verbose=verbose)
            if do_knn:
                wted_dist_lst.append((1 - wt) * mario.dist['ovlp'] + wt * mario.dist['all'])
            print('Filtering bad matched pairs...', flush=True)
            _ = mario.filter_bad_matches(
                'wted', n_clusters, n_components_filter, bad_prop,
                max_iter=max_iter_filter, tol=1e-4, verbose=verbose
            )
            matching_per_batch.append(mario.matching['final'])

        print('Stitching batch-wise matchings together...', flush=True)
        final_matching_lst.append(
            match_utils.stitch(
                perm_lst, [matching_per_batch[j] for j in range(n_batches)], df1.shape[0]
            )
        )
        if do_knn:
            print('Do knn matching...', flush=True)
            dist_argsort = np.array(match_utils.stitch(perm_lst, wted_dist_lst, df1.shape[0]))
            dist_argsort = np.argsort(dist_argsort, axis=1)
            dist_argsort = dist_argsort[:, :knn]
            knn_matching_lst.append([dist_argsort[i, :].tolist() for i in range(df1.shape[0])])

    print('Matching is done, calculating joint embeddings...', flush=True)
    # align all the datasets
    data_aligned_lst = [[] for _ in range(len(data_lst))]
    for i in range(df1.shape[0]):
        if any([len(final_matching_lst[j][i]) == 0 for j in range(len(data_lst))]):
            continue
        for j in range(len(data_lst)):
            if len(final_matching_lst[j][i]) == 1:
                data_aligned_lst[j].append(data_lst[j].iloc[final_matching_lst[j][i][0], :])
            else:
                data_aligned_lst[j].append(np.mean(data_lst[j].iloc[final_matching_lst[j][i], :], axis=0))
    data_aligned_lst = [np.array(each) for each in data_aligned_lst]

    if len(data_lst) == 2:
        _, cca = embed.get_cancor(data_aligned_lst[0], data_aligned_lst[1], embed_dim, max_iter=1000)
        df1_cca, df2_cca = cca.transform(data_aligned_lst[0], data_aligned_lst[1])
        embedding_lst = [df1_cca, df2_cca]
    else:
        embedding_lst = embed.gcca(
            data_aligned_lst, embed_dim, normalization=False, max_iter=max_iter_embed, tol=1e-3, verbose=False
        )

    if normalize_embed:
        for idx in range(len(embedding_lst)):
            embedding_lst[idx] = utils.normalize(embedding_lst[idx])

    print('Saving the results...', flush=True)
    os.makedirs(save_path, exist_ok=True)

    def format_matching(matching):
        res = []
        for matched in matching:
            if len(matched) == 0:
                res.append('')
            else:
                matched = [str(i) for i in matched]
                res.append(','.join(matched))
        return res

    matching_colnames = ['row_idx_' + str(i) for i in range(len(data_lst))]
    final_matching_df = np.array([format_matching(matching) for matching in final_matching_lst]).T
    final_matching_df = pd.DataFrame(final_matching_df, columns=matching_colnames)
    final_matching_df.to_csv(save_path + '/final_matching.csv', index=False)

    if do_knn:
        knn_matching_df = np.array([format_matching(matching) for matching in knn_matching_lst]).T
        knn_matching_df = pd.DataFrame(knn_matching_df, columns=matching_colnames)
        knn_matching_df.to_csv(save_path + '/knn_matching.csv', index=False)

    embed_colnames = ['CCA_' + str(i) for i in range(embed_dim)]
    for i in range(len(data_lst)):
        curr_embedding = embedding_lst[i]
        curr_embedding = pd.DataFrame(curr_embedding, columns=embed_colnames)
        curr_embedding.to_csv(save_path + '/embedding_' + str(i) + '.csv', index=False)

    print('Done!', flush=True)
    if do_knn:
        return final_matching_lst, knn_matching_lst, embedding_lst
    else:
        return final_matching_lst, embedding_lst


import numpy as np
import itertools
from sklearn.utils.extmath import randomized_svd
from sklearn.cluster import KMeans


def align_clusters(z1, z2, K, comb_search=False):
    """Align two cluster labels so that the Hamming distance is minimized.

    Parameters
    ----------
    z1 : array-like of shape (n_samples, )
        First cluster label vector.
    z2 : array-like of shape (n_samples, )
        Second cluster label vector.
    K : int
        Number of clusters, cluster labels are encodes from 0, 1, ..., K-1
    comb_search : bool, default=False
        If False, the function is implemented according
        to Lemma 19 of Gao, C., Ma, Z., Zhang, A. Y., & Zhou, H. H. (2017).
        "Achieving optimal misclassification proportion in
        stochastic block models." JMLR.
        In this case, the function will succeed only if
        there exists a constant C>=1 such that
        1. the minimum cluster size of z1 and z2 are both >= n/(C*K);
        2. after alignment, the Hamming distance between z1 and z2
            is < 1/(C*K).
        Otherwise there is no guarantee that such an alignment can be done
        If True, then this function tries out
        all possible permutations of size K and pick the best one.

    Returns
    -------
    z1 : array-like of shape (n_samples, )
        First cluster label vector.
    z2_aligned : array-like of shape (n_samples, )
        Second cluster label vector after alignment.
    perm: list
        A permutation so that z2_aligned=[perm[k] for k in z2].
    """
    n = len(z1)
    assert n == len(z2)

    z1 = np.array(z1, dtype=int)
    z2 = np.array(z2, dtype=int)

    if comb_search:
        best_perm = None
        best_loss = 2
        for perm in itertools.permutations(np.arange(K)):
            z2_permuted = [perm[k] for k in z2]
            loss = np.mean(z1 != z2_permuted)
            if loss < best_loss:
                best_perm = perm
                best_loss = loss
        if best_perm is not None:
            z2_aligned = [best_perm[k] for k in z2]
            return z1, z2_aligned, best_perm
        else:
            raise(ValueError("Alignment failed!"))

    # don't do combinatorial search

    nbhd_z1 = [np.nonzero(z1 == k) for k in range(K)]
    nbhd_z2 = [np.nonzero(z2 == k) for k in range(K)]

    perm = []

    for i in range(K):
        intersec_sizes = [len(np.intersect1d(nbhd_z1[j], nbhd_z2[i]))
                          for j in range(K)]
        perm.append(np.argmax(intersec_sizes))

    z2_aligned = np.array([perm[k] for k in z2], dtype=int)

    return z1, z2_aligned, perm


def spectral_clustering(X, n_clusters=10, need_svd=True):
    """Do spectral clustering on X.

    Parameters
    ----------
    X : array-like of shape (n_samples, n_features)
        Dataset to do clustering.
    n_clusters: int, default=10
        Number of clusters.
    need_svd: bool, default=True
        If False, then directly do a k-means on rows of X.
        If True, then this function
        (1) computes the truncated SVD of X = U @ diag(s) @ Vh;
        (2) do a k-means clustering on rows of U @ diag(s).

    Returns
    -------
    cluster_label : array-like of shape (n_samples, )
        Cluster label vector.
    """
    if need_svd:
        U, s, _ = randomized_svd(X, n_components=n_clusters)
        X = U @ np.diag(s)

    kmeans = KMeans(n_clusters=n_clusters).fit(X)
    return kmeans.labels_


def compute_centroids(X, cluster_labels, n_clusters):
    """Given cluster labels, compute the centroids of each cluster.

    Parameters
    ----------
    X : array-like of shape(n_samples, n_features)
        Dataset for computation of centroids.
    cluster_labels : array-like of shape (n_samples, )
        Cluster label vector, each entry takes value in {0, ..., n_clusters-1}
    n_clusters : int
        Number of clusters.

    Returns
    -------
    array-like of shape (n_clusters, n_features)
        Each row is a centroid vector.
    """
    cluster_idx = [np.nonzero(cluster_labels == k) for k in range(n_clusters)]
    centroids = [np.mean(X[cluster_idx[k], :], axis=1).squeeze() for k in range(n_clusters)]
    return np.array(centroids)


def jr_kmeans_one_step(X, init_cluster_labels, n_clusters, mismatch_prob=0.1):
    """Do one step of jointly regularized k-means update.

    Given initial cluster labels, say z1, ..., zL, this function computes the initial centroids
    and solves for the next z1, ..., zL, zstar such that the objective function
    sum_{l} sum_{i} 0.5 * || X[l, i, :] - centroids[l, zl[i], :] ||^2
    + log[(1-mismatch_prob)/(mismatch_prob/(n_clusters-1))] * indicator(zl[i] != zstar[i])
    is minimized.

    Parameters
    ----------
    X : list
        A length L list of numpy arrays, where L is the number of modalities and the l-th element in this list
        is of shape (n_samples, n_features_of_modality_l).
    init_cluster_labels : list
        A length L list of integer arrays, encoding the preliminary cluster labels for each modality.
    n_clusters : int
        Number of clusters desired.
    mismatch_prob : float, default=0.1
        A number strictly in (0, 0.5) that determines the strength of regularization.

    Returns
    -------
    ind_cluster_labels : array-like of shape (L, n_samples)
        Each row is an integer vector encoding the clustering label for one modality.
    glob_cluster_labels : array-like of shape (n_samples, )
        An integer vector encoding the clustering label for the global structure.
    final_loss: float
        Value of the loss function.
    """
    if mismatch_prob < 1e-8:
        raise NotImplementedError("mismatch_prob=0, not implemented")

    L = len(X)
    n = X[0].shape[0]

    centroids = [compute_centroids(X[layer],
                                   init_cluster_labels[layer],
                                   n_clusters)
                 for layer in range(L)]

    ind_cluster_labels = np.zeros((L, n), dtype=np.int64)
    glob_cluster_labels = np.zeros(n, dtype=np.int64)

    reg_param = np.log((mismatch_prob/(n_clusters-1))/(1-mismatch_prob))

    final_loss = 0
    for i in range(n):
        best_loss = np.Inf
        best_Z = None
        best_zstar = None

        # suppose zstar_i = k
        for k in range(n_clusters):

            loss_k = np.zeros((n_clusters, L))
            # loss_k is organized as follows
            # -------------------------------
            # l         | 1  2  3           L
            # zl[i]=0   |
            # zl[i]=1   |
            # ...
            # zl[i]=    |

            for layer in range(L):
                loss_k[:, layer] = [0.5 * np.sum((X[layer][i, :]
                                                  - centroids[layer][k, :])**2)
                                    for k in range(n_clusters)]
                # less loss when zl[i] = k
                loss_k[k, layer] = loss_k[k, layer] + reg_param

            # Z is a length L list encoding
            # the estimator for zl[i] when zstar[i] = k
            Z = np.argmin(loss_k, axis=0)

            # calculate the loss w.r.t. Z
            loss_k = [loss_k[Z[layer], layer] for layer in range(L)]
            loss_k = np.sum(loss_k)

            if loss_k < best_loss:
                best_loss = loss_k
                best_Z = Z
                best_zstar = k

        assert best_Z is not None
        assert best_zstar is not None

        ind_cluster_labels[:, i] = np.array(best_Z)
        glob_cluster_labels[i] = best_zstar
        final_loss = final_loss + best_loss / n

    return ind_cluster_labels, glob_cluster_labels, final_loss


def jr_kmeans(X, init_cluster_labels, n_clusters,
              mismatch_prob=0.1, max_iter=20, tol=1e-5, verbose=False):
    """Training jointly regularized k-means clustering using alternating minimization.

    The objective function is
    sum_{l} sum_{i} 0.5* || X[l, i, :] - centroids[l, zl[i], :] ||^2
    + log[(1-mismatch_prob/(n_clusters-1))/mismatch_prob] * indicator(zl[i] != zstar[i]).

    Parameters
    ----------
    X : list
        A length L list of numpy arrays, where L is the number of modalities and the l-th element in this list
        is of shape (n_samples, n_features_of_modality_l).
    init_cluster_labels : list
        A length L list of integer arrays, encoding the preliminary cluster labels for each modality.
    n_clusters : int
        Number of clusters desired.
    mismatch_prob : float, default=0.1
        A number strictly in (0, 0.5) that determines the strength of regularization.
    max_iter : float, default=20
        Maximum number of iterations.
    tol : float, default=1e-5
        Convergence criterion.
    verbose : bool, default=True
        If True, print the progress of each iteration.

    Returns
    -------
    ind_cluster_labels : array-like of shape (L, n_samples)
        Each row is an integer vector encoding the clustering label for one modality.
    glob_cluster_labels : array-like of shape (n_samples, )
        An integer vector encoding the clustering label for the global structure.
    centroids : list
        A length L list, each is an array of shape (n_clusters, n_features_of_modality_l)
        representing the centroids for the l-th modality.
    """
    if mismatch_prob < 1e-8:
        raise NotImplementedError("mismatch_prob=0, not implemented")

    L = len(X)
    assert L == len(init_cluster_labels)
    n_vec = [X[layer].shape[0] for layer in range(L)]
    n = n_vec[0]
    assert np.sum(np.array([(n_each == n) for n_each in n_vec])) == len(n_vec)

    # prev_glob_cluster_labels = init_cluster_labels[0]
    prev_loss = np.Inf
    for ii in range(max_iter):
        ind_cluster_labels, glob_cluster_labels, loss = \
            jr_kmeans_one_step(X, init_cluster_labels,
                               n_clusters, mismatch_prob)

        if verbose:
            print("Now at iteration {}, "
                  "current loss is {}.".format(ii, loss), flush=True)

        if np.abs(prev_loss - loss) < tol:
            centroids = [compute_centroids(
                X[layer], init_cluster_labels[layer], n_clusters
            ) for layer in range(L)]

            return ind_cluster_labels, glob_cluster_labels, centroids

        prev_loss = loss
        init_cluster_labels = ind_cluster_labels

    centroids = [compute_centroids(
        X[layer], init_cluster_labels[layer], n_clusters
    ) for layer in range(L)]

    return ind_cluster_labels, glob_cluster_labels, centroids


import numpy as np


def normalize(df):
    """Center each column and scale each column to have unit standard deviation.

    Parameters
    ----------
    df : array-like of shape (n_samples, n_features)
        Two dimensional array to be normalized.
    Returns
    -------
    df : array-like of shape (n_samples, n_features)
        Two dimensional array after normalization.
    """
    df = df - np.mean(df, axis=0)
    df = df / np.std(df, axis=0)
    return df


def log_normalize(X, scale_factor=1e6):
    """Log-normalize the data.

    Parameters
    ----------
    X: array-like of shape (n_samples, n_features)
        Two dimensional array to normalize.
    scale_factor: float, default=1e6
        Multiple all entries of X by scale_factor before taking log.

    Returns
    -------
    X: array-like of shape (n_samples, n_features)
        Two dimensional array after normalization.
    """
    row_cnts = np.sum(X, axis=1)
    X = ((X * scale_factor).T / row_cnts).T
    X = np.log(X + 1)
    return X


def cdist_correlation(X, Y):
    """Calculate pair-wise Pearson correlation between X and Y.

    Parameters
    ----------
    X: array-like of shape (n_samples_of_X, n_features)
        First dataset.
    Y: array-like of shape (n_samples_of_Y, n_features)
        Second dataset.

    Returns
    -------
    array-like of shape (n_samples_of_X, n_samples_of_Y)
        The (i, j)-th entry is the Pearson correlation between i-th row of X and j-th row of Y.
    """
    n, p = X.shape
    m, p2 = Y.shape
    assert p2 == p

    X = (X.T - np.mean(X, axis=1)).T
    Y = (Y.T - np.mean(Y, axis=1)).T

    X = (X.T / np.sqrt(np.sum(X ** 2, axis=1))).T
    Y = (Y.T / np.sqrt(np.sum(Y ** 2, axis=1))).T

    return 1 - X @ Y.T


def check_min_dist(dist_mat, min_dist):
    """Make sure min(dist_mat) >= min_dist.

    Parameters
    ----------
    dist_mat : array-like of shape (n_samples_1, n_samples_2)
        A two-dimensional distance matrix.
    min_dist : flaot
        Desired minimum distance.

    Returns
    -------
    dist_mat : array-like of shape (n_samples_1, n_samples_2)
        A two-dimensional distance matrix with min(dist_mat) >= min_dist.
    """
    current_min_dist = np.min(dist_mat)
    if current_min_dist < min_dist:
        dist_mat = dist_mat - current_min_dist + min_dist

    return dist_mat


import numpy as np
from sklearn.cross_decomposition import CCA
from . import utils


def get_cancor(X, Y, n_components=10, max_iter=2000):
    """Fit CCA and calculate the canonical correlations.

    Parameters
    ----------
    X: array-like of shape (n_samples, n_features_of_X)
        First dataset.
    Y: array-like of shape (n_samples, n_features_of_Y)
        Second dataset.
    n_components: int, default=2
        Number of CCA components to calculate; must be <= min(n_features_of_X, n_features_of_Y)
    max_iter: int, default=1000
        Maximum number of iterations.

    Returns
    -------
    cancor: array-like of shape (n_components, )
        Vector of canonical components.
    cca: CCA
        CCA object.
    """
    assert X.shape[1] >= n_components
    assert Y.shape[1] >= n_components

    cca = CCA(n_components=n_components, max_iter=max_iter)
    cca.fit(X, Y)
    X_c, Y_c = cca.transform(X, Y)
    cancor = np.corrcoef(
        X_c, Y_c, rowvar=False).diagonal(
        offset=cca.n_components)
    return cancor, cca


def gcca_init(data_list, n_components=10, normalization=True):
    """Initialization of GCCA.
    Parameters
    ----------
    data_list: list
        A list of data matrices, e.g., [X1, X2, X3, X4]
    n_components : int, default=10
        Number of GCCA components.
    normalization : bool, default=True
        If true, center each column and scale each column to have unit standard deviation.

    Returns
    ----------
    score_list: list
        A list of projected data.
    """

    # normalize
    if normalization:
        for data in data_list:
            data = utils.normalize(data)
    # run CCA on the first two data matrices
    _, cca = get_cancor(data_list[0], data_list[1], n_components=n_components)
    score_list = list(cca.fit_transform(data_list[0], data_list[1]))
    avg_score = (score_list[0] + score_list[1]) / 2
    # run OLS regressions for the rest of data matrices
    for data in data_list[2:]:
        B = np.linalg.lstsq(data, avg_score, rcond=None)[0]
        score_list.append(data @ B)

    return score_list


def gcca_refine(data_list, init_scores, max_iter=500, tol=1e-3, verbose=True):
    """Refinement of GCCA.

    Parameters
    ---------
    data_list : list
        A list of data matrices.
    init_scores : list
        A list of initial scores.
    max_iter : int, default=500
        Number of maximum iterations.
    tol : float, default=1e-3
        Algorithm terminates when the change of objective function value is <=tol.
    verbose : bool, default=True
        Print details if True.

    Returns
    ----------
    all_scores : list
        A list of projected data.
    """

    m = len(init_scores)
    if len(init_scores[0]) == 0:
        n_components = 1
        n = len(init_scores[0])
    else:
        n, n_components = init_scores[0].shape

    all_scores = np.array([np.empty_like(init_scores[0])] * m)
    for ii in range(n_components):
        if verbose:
            print("Computing the {}-th canonical score...".format(ii), flush=True)
        # regress out the influence of all_score[:, :, :ii]
        if ii > 0:
            curr_data_list = []
            for idx, data in enumerate(data_list):
                coef = np.linalg.lstsq(all_scores[idx, :, :ii], data, rcond=None)[0]
                curr_data_list.append(data - all_scores[idx, :, :ii] @ coef)
        else:
            curr_data_list = data_list

        # preparations for entering main iteration
        if n_components == 1:
            curr_scores = [score for score in init_scores]
        else:
            curr_scores = [score[:, ii] for score in init_scores]
        score_sum = sum(score for score in curr_scores)
        prev_obj = np.inf

        # alternating minimization
        for iter_idx in range(max_iter):
            for jj in range(m):
                y = (score_sum - curr_scores[jj]) / (m - 1)
                X = curr_data_list[jj]
                coef = np.linalg.lstsq(X, y, rcond=None)[0]
                coef = coef / np.sqrt(np.sum(coef ** 2))
                score = X @ coef
                score_sum = score_sum - curr_scores[jj] + score
                curr_scores[jj] = score
            # check convergence
            obj = 0
            for jj in range(m - 1):
                obj += np.sqrt(np.sum((curr_scores[jj] - curr_scores[jj + 1]) ** 2))

            if verbose and iter_idx % 50 == 0:
                print("At iteration {}, the objective value is {}.".format(iter_idx, obj), flush=True)

            if abs(obj - prev_obj) < tol:
                break
            else:
                prev_obj = obj

        if verbose:
            print("Finished computing the {}-th canonical score, the objective is {}.".format(ii, obj), flush=True)

        for jj in range(m):
            if n_components == 1:
                all_scores[jj, :] = curr_scores[jj]
            else:
                all_scores[jj, :, ii] = curr_scores[jj]

        # curr_scores are linear combinations of the residual matrices
        # convert them into linear combinations of the original data matrices
    #         for jj in range(m):
    #             coef = np.linalg.lstsq(data_list[jj], curr_scores[jj], rcond=None)[0]
    #             if n_components == 1:
    #                 all_scores[jj, :] = data_list[jj] @ coef
    #             else:
    #                 all_scores[jj, :, ii] = data_list[jj] @ coef

    return [all_scores[jj, :, :] for jj in range(m)]


def gcca(data_list, n_components=10, normalization=True, max_iter=500, tol=1e-3, verbose=True):
    """Embedding multiple datasets via generalized canonical correlation analysis.
    Parameters
    ---------
    data_list : list
        A list of data matrices.
    n_components : int, default=10
        Number of GCCA components.
    normalization : bool, default=True
        If true, center each column and scale each column to have unit standard deviation.
    max_iter : int, default=500
        Number of maximum iterations.
    tol : float, default=1e-3
        Terminate the algorithm when the change in objective value is <=tol.
    verbose : bool, default=True
        Print details if True.

    Returns
    ----------
    list
        A list of projected data.
    """
    init_scores = gcca_init(data_list, n_components, normalization)
    return gcca_refine(data_list, init_scores, max_iter, tol, verbose)